use crate::prelude::*;
use anyhow::Result;
use cgmath::{Deg, Matrix4, Vector2, Vector3, num_traits::clamp, vec3};

pub struct TopDownCameraControl {
    zoom_levels: Vec<f32>,
    current_zoom_level: usize,

    rotation: Deg<f32>,
    arc: Deg<f32>,

    mouse_position: (u32, u32),
    stick_direction: Vector2<f32>,

    rotation_start: Deg<f32>,
    arc_start: Deg<f32>,
    mouse_position_start: Option<(u32, u32)>,
}

impl TopDownCameraControl {
    const SCALE: f32 = 0.3;
    const MIN_PITCH: f32 = 10.0;
    const MAX_PITCH: f32 = 80.0;

    pub fn new(view: &mut View) -> Result<Self> {
        let me = Self {
            zoom_levels: vec![
                15.0, 14.0, 13.0, 12.0, 11.0, 10.0, 9.0, 8.0, 7.0, 6.0, 5.0, 4.0, 3.0, 2.0,
            ],
            current_zoom_level: 10,

            rotation: Deg(0.0),
            arc: Deg(50.0),

            mouse_position: (0, 0),
            stick_direction: Vector2::zero(),

            rotation_start: Deg(0.0),
            arc_start: Deg(0.0),
            mouse_position_start: None,
        };

        me.set_camera_offset(view)?;

        Ok(me)
    }

    pub fn set_zoom_levels(
        &mut self,
        zoom_levels: Vec<f32>,
        current_zoom_level: usize,
        view: &mut View,
    ) -> Result<()> {
        self.current_zoom_level = current_zoom_level.min(zoom_levels.len() - 1);
        self.zoom_levels = zoom_levels;

        self.set_camera_offset(view)
    }

    fn set_camera_offset(&self, view: &mut View) -> Result<()> {
        view.camera_mut()
            .set_eye_offset(self.calculate_camera_offset());
        view.update_buffer()
    }

    fn calculate_camera_offset(&self) -> Vector3<f32> {
        let dir = vec3(0.0, -self.zoom_levels[self.current_zoom_level], 0.0);

        let offset = Matrix4::from_angle_z(self.rotation)
            * Matrix4::from_angle_x(-self.arc)
            * dir.extend(1.0);

        offset.truncate()
    }

    pub fn hold(&mut self) {
        self.mouse_position_start = Some(self.mouse_position);
        self.rotation_start = self.rotation;
        self.arc_start = self.arc;
    }

    pub fn release(&mut self) {
        self.mouse_position_start = None;
    }

    pub fn stick_direction(&mut self, stick: Vector2<f32>) {
        self.stick_direction = stick;
    }

    pub fn stick_movement(&mut self, factors: Vector2<f32>, view: &mut View) -> Result<()> {
        self.rotation = Deg((self.rotation.0 + (self.stick_direction.x * factors.x)) % 360.0);
        self.arc = Deg(clamp(
            self.arc.0 - (self.stick_direction.y * factors.y),
            TopDownCameraControl::MIN_PITCH,
            TopDownCameraControl::MAX_PITCH,
        ));

        self.set_camera_offset(view)
    }

    pub fn mouse_move(&mut self, x: u32, y: u32, view: &mut View) -> Result<()> {
        self.mouse_position = (x, y);

        if let Some((start_x, start_y)) = self.mouse_position_start {
            let x_diff = (start_x as i32 - x as i32) as f32 * TopDownCameraControl::SCALE;
            let y_diff = (start_y as i32 - y as i32) as f32 * TopDownCameraControl::SCALE;

            self.rotation = Deg((self.rotation_start.0 + x_diff) % 360.0);
            self.arc = Deg(clamp(
                self.arc_start.0 - y_diff,
                TopDownCameraControl::MIN_PITCH,
                TopDownCameraControl::MAX_PITCH,
            ));

            self.set_camera_offset(view)?;
        }

        Ok(())
    }

    pub fn scroll_wheel(
        &mut self,
        y: i32,
        direction: MouseWheelDirection,
        view: &mut View,
    ) -> Result<()> {
        match direction {
            MouseWheelDirection::Flipped => {
                if y < 0 {
                    self.current_zoom_level =
                        (self.current_zoom_level + (-y as usize)).min(self.zoom_levels.len() - 1);
                } else {
                    if self.current_zoom_level >= y as usize {
                        self.current_zoom_level -= y as usize;
                    }
                }
            }

            _ => {
                if y < 0 {
                    if self.current_zoom_level >= (-y as usize) {
                        self.current_zoom_level -= -y as usize;
                    }
                } else {
                    self.current_zoom_level =
                        (self.current_zoom_level + y as usize).min(self.zoom_levels.len() - 1);
                }
            }
        }

        self.set_camera_offset(view)
    }
}
