pub mod deferred;
mod lights;
mod pipelines;
pub mod rasterizershader;
pub mod traditional;

use super::shared::{bufferhandler::*, shadercompiler::Shaders};

use crate::prelude::*;
use anyhow::Result;
use asset::RasterizerDescriptorLayouts as AssetLayouts;
use pipelines::RasterizerPipelines;
use utilities::{impl_reprc, prelude::cgmath::Vector3};

use std::sync::{Arc, Mutex};

impl_reprc!(
    #[derive(Copy, Debug)]
    pub struct SceneInfo {
        #[assume_reprc]
        point_light_count: u32,
        #[assume_reprc]
        directional_light_count: u32,
        #[assume_reprc]
        spot_light_count: u32,
        #[assume_reprc]
        view_position: Vector3<f32>,
    }
);

#[macro_export]
macro_rules! update_iter {
    ( $handler: ident, $(($name: ident, $primitive_type: ty),)+ ) => {
        paste::expr! {
            $(
                let [<$name _iter>] = {
                    Box::new(move
                        |
                            renderer: &SafeCommandBuffer<$primitive_type>,
                            recorder: &mut CommandBufferRecorder<'_>,
                            primitive: &PrimitiveData<$primitive_type>,
                            additional_sets: &[&Arc<DescriptorSet>],
                            view: &View,
                        | {
                            let mut desc_sets: Vec<&Arc<DescriptorSet>> =
                                vec![&primitive.descriptor_set, view.descriptor_set()];

                            for set in additional_sets {
                                desc_sets.push(*set);
                            }

                            renderer.render(recorder, desc_sets.as_slice(), &primitive.vertex_buffer)?;

                            Ok(())
                        }
                    )
                };

                $handler.[<set_ $name _function>]([<$name _iter>]);
            )+
        }
    };
}

pub fn create_shadow_buffer_handler(
    device: &Arc<Device>,
    queue: &Arc<Mutex<Queue>>,
    image_count: usize,
    render_pass: Arc<RenderPass>,
    shader: &Shaders,
) -> Result<BufferHandler> {
    let globals_layout = View::desc_layout(device)?;
    let model_layout = AssetLayouts::descriptor_layout(device.clone())?;
    let model_bone_layout = AssetLayouts::bone_descriptor_layout(device.clone())?;
    let model_color_layout = AssetLayouts::color_descriptor_layout(device.clone())?;
    let model_color_bone_layout = AssetLayouts::color_bone_descriptor_layout(device.clone())?;
    let model_color_normal_layout = AssetLayouts::color_normal_descriptor_layout(device.clone())?;
    let model_color_normal_bone_layout =
        AssetLayouts::color_bone_normal_descriptor_layout(device.clone())?;

    let light_info_layout = DescriptorSetLayout::builder()
        .add_layout_binding(
            0,
            VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
            VK_SHADER_STAGE_FRAGMENT_BIT,
            0,
        )
        .add_layout_binding(
            1,
            VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
            VK_SHADER_STAGE_FRAGMENT_BIT,
            0,
        )
        .build(device.clone())?;

    let attachment_info = [false];

    let create_info = BufferHandlerCreateInfo::new(
        shader,
        device,
        queue,
        image_count,
        BufferHandlerType::graphics(
            &render_pass,
            &attachment_info,
            0,
            RasterizerPipelines::create_shadow_pipeline::<PositionOnly>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![&model_layout, &globals_layout, &light_info_layout],
                push_constants: &[],
            },
            RasterizerPipelines::create_shadow_pipeline::<PositionColorUV>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![&model_color_layout, &globals_layout, &light_info_layout],
                push_constants: &[],
            },
            RasterizerPipelines::create_shadow_pipeline::<PositionNormal>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![&model_layout, &globals_layout, &light_info_layout],
                push_constants: &[],
            },
            RasterizerPipelines::create_shadow_pipeline::<PositionNormalBone>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![&model_bone_layout, &globals_layout, &light_info_layout],
                push_constants: &[],
            },
            RasterizerPipelines::create_shadow_pipeline::<PositionNormalColor>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![&model_layout, &globals_layout, &light_info_layout],
                push_constants: &[],
            },
            RasterizerPipelines::create_shadow_pipeline::<PositionNormalColorUV>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![&model_color_layout, &globals_layout, &light_info_layout],
                push_constants: &[],
            },
            RasterizerPipelines::create_shadow_pipeline::<PositionNormalColorUVBone>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![
                    &model_color_bone_layout,
                    &globals_layout,
                    &light_info_layout,
                ],
                push_constants: &[],
            },
            RasterizerPipelines::create_shadow_pipeline::<PositionNormalColorUVNormalUV>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![
                    &model_color_normal_layout,
                    &globals_layout,
                    &light_info_layout,
                ],
                push_constants: &[],
            },
            RasterizerPipelines::create_shadow_pipeline::<PositionNormalColorUVNormalUVBone>,
            PipelineLayoutInfo {
                descriptor_layouts: vec![
                    &model_color_normal_bone_layout,
                    &globals_layout,
                    &light_info_layout,
                ],
                push_constants: &[],
            },
        ),
    );

    BufferHandler::new(&create_info, false)
}
