use context::prelude::*;
use ecs::World;

use std::{
    sync::{
        Arc,
        mpsc::{Receiver, channel},
    },
    thread,
};

use anyhow::{Result, anyhow};

pub struct LoadingScreen<T, G>
where
    T: Send + Sync + 'static,
    G: TopLevelGui + TopGui + Send + Sync + 'static,
{
    result: Option<T>,
    receiver: Receiver<T>,
    gui: Arc<G>,
}

impl<T, G> LoadingScreen<T, G>
where
    T: Send + Sync + 'static,
    G: TopLevelGui + TopGui + Send + Sync + 'static,
{
    pub fn load<R, L>(world: &mut World, gui: Arc<G>, loader: L) -> Result<Self>
    where
        R: Fn(T) -> Result<()> + Send + Sync + 'static,
        T: Send + Sync + 'static,
        L: FnOnce() -> T + Send + Sync + 'static,
    {
        gui.enable(world)?;

        let (sender, receiver) = channel();

        thread::spawn(move || {
            let _ = sender.send(loader());
        });

        Ok(Self {
            result: None,
            receiver,
            gui,
        })
    }

    pub fn check_ready(&mut self) -> bool {
        if self.result.is_some() {
            return true;
        }

        match self.receiver.try_recv() {
            Ok(result) => {
                self.result = Some(result);
                true
            }
            Err(_) => false,
        }
    }

    pub fn take_result(mut self, world: &mut World) -> Result<T> {
        self.gui.disable(world)?;
        self.result
            .take()
            .ok_or(anyhow!("missing loading screen payload!"))
    }
}
